<!doctype html>
<html>
	<head>
		<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
		<link rel="icon" href="{{ url_for('static', filename='site.ico') }}" type="image/vnd.microsoft.icon" />
		<link rel="shortcut icon" href="{{ url_for('static', filename='site.ico') }}" type="image/vnd.microsoft.icon" />
		<link rel="stylesheet" type="text/css" href="{{ loc.bootstrap_css }}" crossorigin="anonymous" />
		<link rel="stylesheet" type="text/css" href="static/site.css" />

		<title>{{title}}</title>
	</head>
	<body>

		<header>
			<div id="df_alerts">
				<div id="init_alert" class="alert alert-primary" role="alert">
					Loading. Please wait.
				</div>
			</div>
			<div id="df_spinner" class="spinner-grow text-primary" role="status">
				<span class="sr-only">
					Loading…
				</span>
			</div>

			<div class="modal fade" id="df_modal" tabindex="-1" role="dialog" aria-hidden="true" aria-labelledby="df_modal_title" data-backdrop="static">
				<div class="modal-dialog modal-dialog-centered" role="document">
					<div class="modal-content">
						<div class="modal-header">
							<h5 id="df_modal_title" class="modal-title" id="exampleModalCenterTitle">
								Set a title
							</h5>
						</div>
						<div id="df_modal_body" class="modal-body">
							<p>
								You didn't set the modal content.
							</p>
						</div>
						<div id="df_modal_footer" class="modal-footer">
							You forgot to add buttons.
						</div>
					</div>
				</div>
			</div>
		</header>

		<main>
			<div id="df_header">
				<h1 class="title">{{title}}</h1>
			</div>
			<div id="df_main" class="container-fluid">
				<noscript>
					<p>
						You disabled Javascript.
						<br />
						Sorry, but this site won't work without it!
						<br />
						(We promise not to do anything nefarious.)
					</p>
				</noscript>
			</div>
		</main>

		<footer>
			<div id="df_footer">
				<div id="df_footer_left" class="left">
					&nbsp;
				</div>
				<div id="df_footer_right" class="right">
					&nbsp;
				</div>
			</div>
		</footer>

		<script type="text/javascript" src="{{ loc.jquery }}"></script>
		<script>
			window.deframed_version = "{{version}}";
			window.deframed_debug = "{{debug}}";
			$("#df_main").html("<p>Content will load shortly.</p>");
		</script>
		<script type="text/javascript" src="{{ loc.poppler }}"crossorigin="anonymous"></script>
		<script type="text/javascript" src="{{ loc.bootstrap_js }}"crossorigin="anonymous"></script>
		<script type="text/javascript" src="{{ loc.msgpack }}"crossorigin="anonymous"></script>
		<script type="text/javascript" src="{{ loc.mustache }}"crossorigin="anonymous"></script>
		<script type="text/javascript" src="{{ loc.plotly }}"crossorigin="anonymous"></script>
		<script type="text/javascript" src="static/main.js"></script>
	</body>
</html>
